/*
 *Cisco Vlab Proxy server
 */

#include <sys/types.h>
#include <sys/param.h>

#include <sys/socket.h>
#include <sys/wait.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>

#include "proxy.h"

#define PRX_HOLD_CONN 10
#define DEFAULT_LISTEN_PORT 4000


/* logging into file*/

FILE* logfile;
//typedef enum loglevel_t {INFO, WARN, ERR} logtype;

pthread_mutex_t logmutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t sqlmutex = PTHREAD_MUTEX_INITIALIZER;

void usage(void);
int main(int, char**);
void writelog(const char*);
char *  tstamp(char*);
//void shout(logtype,const char*);

int auth_flag;
int verbose_flag;
int debug_flag;

/*store Listen network options here*/
struct sockaddr_in prx_addr;

/*Main proxy fd*/
int prx_sockfd;

void
writelog(const char* message)
{
	char* stamp;

	stamp=(char *)malloc(100*sizeof(char));
	stamp=tstamp(stamp);

	fprintf(stderr,"shout:%s: %s\n",stamp,message);

	if(logfile!=NULL) 
	{
		if(debug_flag)
			fprintf(stderr,"locking log\n");

		pthread_mutex_lock( &logmutex );
		
		fprintf(logfile,"%s %s\n",stamp, message); 
		fflush(logfile);

		if(debug_flag)
			fprintf(stderr,"unlocking log\n");

		pthread_mutex_unlock( &logmutex );
	}
	
	free(stamp);
	stamp=NULL;
}

void
shout(logtype loglevel, const char* message)
{
	if((verbose_flag && loglevel>INFO) || debug_flag)
		writelog(message);	
}

char*
tstamp(char* result)
{
	struct tm *tm_ptr;
	time_t the_time;

	//char buf[256];
//	char* result;

	(void) time(&the_time);
	tm_ptr = localtime(&the_time);

	//strftime(buf,256,"%A %d %B, %l:%S %p",tm_ptr);
	//printf("tstamp:%s\n",buf);

	strftime(result,100,"%s %H:%M.%S %d.%m.%Y",tm_ptr);

	return result;
}


int
main (int argc, char **argv)
{
	char ch;
	int prx_port=DEFAULT_LISTEN_PORT; /*Port where proxy listen for client connections*/
	size_t bits;
	socklen_t cn_size;
	char* path;  // path to logfile
	struct thread_arg *thread_info;
	logfile=NULL;
	
	while ((ch = getopt(argc, argv, "avdmhp:i:l:")) != -1) {
		switch (ch) {
		case 'a':
			auth_flag = 1;
			break;

		case 'v':
			verbose_flag = 1;
			break;	

		case 'd':
			debug_flag = 1;
			break;

		case 'p':
			prx_port = atoi(optarg);
			if ((prx_port > 65535) || (prx_port < 1024))
						    err(EXIT_FAILURE,
							"Use port number between 1024 and 65535.\n");
			break;

		case 'i':
		#ifdef bsd
			bits=inet_net_pton(AF_INET,
			    optarg,
			    &prx_addr.sin_addr,
			    sizeof(prx_addr.sin_addr));
		#endif
		#ifdef linux 
			bits=inet_pton(AF_INET,
			    optarg,
			    &prx_addr.sin_addr);
		#endif

			break;
		case 'l':
			if((path=(char*)malloc((1+strlen(optarg))*sizeof(char)))==NULL)
				err(EXIT_FAILURE,"Not enough memory for path to file\n");
			strncpy(path, optarg,strlen(optarg));
			/* logging to file */
			if((logfile=fopen(path,"a"))==NULL)	{shout(ERR,"cannot open logfile");		};

			free(path);	


			break;

		case 'h':	
		case '?':
		default:
			usage();

		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 1)
		usage();

	shout(ERR,"Vlab proxy started ");
// debugging of mysql
//	char nic[]="auth_token";
//	prx_auth_callback(nic,3000);
//	exit(0);

	/*Prepare prx_addr struct to bind*/
	prx_addr.sin_port = htons(prx_port);
	prx_addr.sin_family = AF_INET;   // host byte order
	memset(prx_addr.sin_zero, '\0', sizeof prx_addr.sin_zero);

	if ((prx_sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		err(EXIT_FAILURE,"Create proxy failed.\n");
	
//	if (setsockopt(prx_sockfd, SOL_SOCKET, SO_REUSEADDR, NULL, 0) == -1)
//		err(EXIT_FAILURE,"Setsockopt proxy failed.\n");
	
	if (bind(prx_sockfd, (struct sockaddr *)&prx_addr, sizeof(prx_addr)) == -1)
		err(EXIT_FAILURE,"Proxy bind at address: %s port: %d failed \n",
		    inet_ntoa(prx_addr.sin_addr), prx_port);
			
	if (listen(prx_sockfd, PRX_HOLD_CONN) == -1) 
		err(EXIT_FAILURE,"Proxy listen at address: %s port: %d failed \n",
		    inet_ntoa(prx_addr.sin_addr), prx_port);

	fprintf(stderr,"Proxy started at address: %s port: %d, \n",inet_ntoa(prx_addr.sin_addr),
	    prx_port);


	/* Create thread per connection after accept */
	while(1){
		/* Allocate thread info struct this struct contains all needed
		   information for thread to proceed our connection.*/
		
		if ((thread_info = malloc (sizeof(struct thread_arg))) == NULL)
			err(ENOMEM,"Not Enough memory to process another request\n");


		cn_size = sizeof(thread_info->th_addr_in);
				
		if ((thread_info->th_sockfd =
			accept(prx_sockfd, (struct sockaddr *)&thread_info->th_addr_in,
			    &cn_size)) == -1) {
			perror("accept");
			continue;
		}

		fprintf(stderr, "Proxy vlab server accepted connection\n");
		
		pthread_create(&thread_info->th_thread, NULL, (void *)prx_client_start,thread_info);
		
	}	

	close(prx_sockfd);
		
	return 0;
	
}

void
usage()
{
	printf("Vlab Proxy server\n");
	printf("Accepting these parameters \n");

	printf("usage: proxy [vdma] [-p port] [-i IP address]\n");
	printf("-v verbose,\n -d debug,\n -m mysql authentication,\n -a use auth,\n"
               "-i IP address,\n -p port number\n");

	exit(EXIT_FAILURE);
	
}
