/*
 *Proxy vlab client side. 
 */


#include <sys/types.h>
#include <sys/param.h>

#include <sys/socket.h>
#include <sys/wait.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <signal.h>
#include <pthread.h>

#include "proxy.h"

#define LINESIZE 512

/*
 * Return values of chkstring Bad* are fatal connection is closed
 * after them immediately.
 */
#define BadMsgHeader  -1
#define BadAuthToken  -2
#define BadPortString -3
#define BadMalloc     -4
#define MsgTooLength  -5

extern int verbose_flag;
extern int debug_flag;

static int chkstring(struct cn_info*, char *, size_t);

static int prx_get_addr_in(struct sockaddr_in *, int);
static int prx_relay_loop(int, int);


/*
 *This function initiate client part of connection get first initial line 
 *with authentication code from client, parse it and check it against mysql.
 */
int
prx_client_start(void *data)
{
	struct thread_arg *info;
	struct cn_info *cn;

	size_t recv_s;
	size_t bytes_read;
	
	char buf[LINESIZE];
	#define MSGLEN 200 
	char message[MSGLEN+1]; 

	int ret;
	
	bytes_read = 0;
		
	info = (struct thread_arg *)data;
	
	if ((cn = malloc(sizeof (struct cn_info))) == NULL)
		goto out;
	
	send(info->th_sockfd,"Proxy vlab at your service\n",28,0);

	char* addr;
	addr=(char *)malloc(16*sizeof(char));
	if(inet_ntop(AF_INET,(struct in_addr *)&(info->th_addr_in.sin_addr),addr,15)==NULL)
		shout(ERR,"error in inet_ntop");


	/* Read one line from connection and parse it. */
	while (bytes_read < LINESIZE) {
		/* Read one character from connection because I want to detect '\n' */
		if ((recv_s=recv(info->th_sockfd,&buf[bytes_read],1,0)) == -1) {
			fprintf(stderr,"Client thread id: %x exited in recv\n",
			    (unsigned int)info->th_thread);
			goto out;
		}
		
		if (buf[bytes_read] == '\n')
			break;
		
		bytes_read++;
	}

	/* echo mode  */
//	if (debug_flag)
//		send(info->th_sockfd,buf,strlen(buf),0);

	/* open socket for relaying */
	if ((cn->cn_sockfd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr,"Client thread id: %x exited in socket.\n",
		    (unsigned int)info->th_thread);
		snprintf(message, MSGLEN, "%s %d %d %s ERR_SOCKET", 
			addr,
			ntohs(info->th_addr_in.sin_port),
			ntohs( cn->cn_addr_in.sin_port),
			cn->cn_auth_token);
		shout(ERR,message);


	}
	
	/* Check received string, and fill cn with values.*/
	if ((ret = chkstring(cn,buf,bytes_read)) ==
	    (BadMsgHeader | BadAuthToken | BadMalloc | BadPortString))
	{   	
		snprintf(message, MSGLEN, "%s %d %d %s ERR_INIT_STRING", 
			addr,
			ntohs(info->th_addr_in.sin_port),
			ntohs( cn->cn_addr_in.sin_port),
			cn->cn_auth_token);
		shout(ERR,message);
		goto out;
	}
//	if (verbose_flag)
		fprintf(stderr,"I received auth_token: '%s', relay_port: '%d'\n",
		    cn->cn_auth_token, cn->cn_relay_port);

	if ((prx_auth_callback(cn->cn_auth_token, cn->cn_relay_port)) < 0){
		free(cn->cn_auth_token);
		fprintf(stderr,"Auth callback failed\n");
		snprintf(message, MSGLEN, "%s %d %d %s ERR_AUTH", 
			addr,
			ntohs(info->th_addr_in.sin_port),
			ntohs( cn->cn_addr_in.sin_port),
			cn->cn_auth_token);
		shout(ERR,message);


		goto out;
	}
	
	if ((prx_get_addr_in(&cn->cn_addr_in, cn->cn_relay_port)) < 0)
		goto out;

	/* Connect to relayed side */
	if (connect(cn->cn_sockfd, (struct sockaddr *)&cn->cn_addr_in,
				sizeof(cn->cn_addr_in)) == -1) {

		fprintf(stderr,"Client thread id: %x exited in connect\n",
				(unsigned int)info->th_thread);
		snprintf(message, MSGLEN, "%s %d %d %s ERR_CONNECT_ON_CLIENTSIDE", 
				addr,
				ntohs(info->th_addr_in.sin_port),
				ntohs( cn->cn_addr_in.sin_port),
				cn->cn_auth_token);
		shout(ERR,message);


		if (send(info->th_sockfd,
					"proxy error: i can't connect on relayed side, giving up\n\r",
					57, 0) ==  -1) {

			fprintf(stderr,"Client thread id: %x exited in send\n",    
					(unsigned int)info->th_thread);
		}	
		goto out;
	}

	snprintf(message, MSGLEN, "%s %d %d %s CONNECT", 
			addr, ntohs(info->th_addr_in.sin_port),ntohs( cn->cn_addr_in.sin_port),cn->cn_auth_token);
	shout(ERR,message);
	free(addr);


	/* relay messages */
	prx_relay_loop(cn->cn_sockfd,info->th_sockfd);



	addr=(char *)malloc(16*sizeof(char));
	if(inet_ntop(AF_INET,(struct in_addr *)&(info->th_addr_in.sin_addr),addr,15)==NULL)
		shout(ERR,"error in inet_ntop");
	snprintf(message, (10+16+6+6+strlen(cn->cn_auth_token)), "%s %d %d %s DISCONNECT", 
		addr, ntohs(info->th_addr_in.sin_port),ntohs( cn->cn_addr_in.sin_port),cn->cn_auth_token);
	shout(ERR,message);
	free(addr);


	
out:	
	close(info->th_sockfd);
	close(cn->cn_sockfd);
	
	free(info);
	free(cn);
	
	pthread_exit(0);
}

/*
 * Check received msg. This function is called only once after connection start to
 * authenticate connection and relaying.
 *
 * XXX implement test for relayMsg part of message.
 */

static int
chkstring(struct cn_info *cn, char *msg, size_t msg_size)
{
	char *msg_pointer;
	
	int header_len;
	int auth_token_len;
	int port_len;
	int ret;
	
	ret = 0;

	header_len = strlen("relay");
	
	/*
	 * Standard message sent with applet is this
	 * relay auth_string port
	 * relay heslooo 2000
	 * Updated: relay auth_string port relayMsg 
	 */

	/* Check for header  */
	if ((strncmp(msg,"relay",header_len) < 0) ||
	    (strcspn(msg," ") > header_len)) {
	    	shout(WARN,"bad init");
		ret = BadMsgHeader;
		goto ret;
	}

	msg_pointer = msg+header_len+1;

	/* Authtoken is divided with ' ' from port. */
	if ((auth_token_len = strcspn(msg_pointer," ")) < 0){
		shout(WARN,"bad auth token");
		ret = BadAuthToken;
		goto ret;
	}

	/* Copy auth_token to per connection info structure. */
	if ((cn->cn_auth_token = malloc(auth_token_len+1 * sizeof(char))) == NULL){
		ret = BadMalloc;
		goto ret;
	}
	
	// strncpy(cn->cn_auth_token,msg_pointer,auth_token_len+1); << adam ? printf(%s) ?
	memset(cn->cn_auth_token,'\0',auth_token_len+1);
	strncpy(cn->cn_auth_token,msg_pointer,auth_token_len);

	msg_pointer +=auth_token_len+1;

	/* find end of port string. */
	if ((port_len = strcspn(msg_pointer," ")) < 0){
		ret = BadPortString;

		free(cn->cn_auth_token);
		cn->cn_auth_token = NULL;
		
		goto ret;
	}

	/* Try to convert port string to number */
	if ((cn->cn_relay_port = (int)strtol(msg_pointer, (char **)NULL, 10)) == 0){
		ret = BadPortString;

		free(cn->cn_auth_token);
		cn->cn_auth_token = NULL;
				
		goto ret;
	}

	/*
	 * This is non fatal error :I got longer message than I can parse.
	 * I already have all info that I need to go.
	 */
	if ((header_len+1+auth_token_len+1+port_len) != strlen(msg)){
		ret = MsgTooLength;

		fprintf(stderr,"Got malformed auth packet, something is wrong..\n");

		goto ret;
	}
	shout(INFO,"relay msg OK");
	
	return 0;

ret:
	return ret;
}


/*
 * This function fill struct sockaddr_in which is needed for applet->router connection.
 *
 * TODO: When we want to support clusters we can fo it here everything else is same for
 * cluster/non-cluster arch.
 */

static int
prx_get_addr_in(struct sockaddr_in *cn_addr, int cn_port)
{

	/* TODO DO we want this as runtime option ? */
#ifndef __VLAB_CLUSTER__
	#define LOCALHOST "127.0.0.1"

	struct hostent *he;

	if ((he=gethostbyname(LOCALHOST)) == NULL) 
		return -1;
		
	cn_addr->sin_family = AF_INET; 
	cn_addr->sin_port = htons(cn_port);
	cn_addr->sin_addr = *((struct in_addr *)he->h_addr);
	
	memset(cn_addr->sin_zero, '\0', sizeof cn_addr->sin_zero);
	
#endif	
	
	return 0;
}

#define BUF_SIZE 32

/*
 * This function loops forever between realyfd nad clientfd
 * and sends  messages between them.
 */
int
prx_relay_loop(int relayfd, int clientfd)
{
	int retval;
	int maxfd;

	int read_bytes;
		
	fd_set rfds;
	struct timeval tv;

	char buf[BUF_SIZE];
		
	for (;;) {

		FD_ZERO(&rfds);

		FD_SET(clientfd, &rfds);
		FD_SET(relayfd, &rfds);
	
		/* Wait up to five seconds. */
		tv.tv_sec = 5;
		tv.tv_usec = 0;

		maxfd= clientfd > relayfd ? clientfd: relayfd;

		/* Select fd with data or closed connection */
		retval = select(maxfd+1, &rfds, NULL , NULL, &tv);

		if (retval) {
			printf("data is available now\n");

			/* Client is sending data to relayed server */
			if(FD_ISSET(clientfd,&rfds)) {
				
				if((read_bytes = recv(clientfd,&buf,BUF_SIZE,0)) < 0)
					return -1;
				
				if (read_bytes > 0) { 
					printf("received %d bytes on clientside\n",(int)read_bytes);

					if ((send(relayfd, &buf, read_bytes, 0)) < 0)
						return -1;
					
				} else {
					printf("socket triggered, but received %d"
					    " bytes on clientside\n",(int)read_bytes);
					
					printf("assuming closed connection, closing serverside too\n");

					return 0;
				}
			}

			/* Relayed server response */
			if(FD_ISSET(relayfd,&rfds)) {
				
				if((read_bytes = recv(relayfd,&buf,BUF_SIZE,0)) < 0)
					return -1;
				
				if (read_bytes > 0) { 
					printf("received %d bytes on clientside\n",(int)read_bytes);

					if ((send(clientfd, &buf, read_bytes, 0)) < 0)
						return -1;
					
				} else {
					printf("socket triggered, but received %d"
					    " bytes on clientside\n",(int)read_bytes);
					
					printf("assuming closed connection, closing serverside too\n");

					return 0;
				}
			} 
		}	
	}

	return 0;
}
